#![allow(dead_code)]

use simconnect_sdk::{
    Condition, DataType, Notification, NotificationData, Period, SimConnect, SimConnectError,
    SimConnectObjectExt,
};

/// A data structure that will be used to receive data from SimConnect.
#[derive(Debug, Clone)]
pub struct GpsData {
    lat: f64,
    lon: f64,
    alt: f64,
}

impl SimConnectObjectExt for GpsData {
    fn register(client: &mut SimConnect, id: u32) -> Result<(), SimConnectError> {
        client.add_to_data_definition(id, "PLANE LATITUDE", "degrees", DataType::Float64)?;
        client.add_to_data_definition(id, "PLANE LONGITUDE", "degrees", DataType::Float64)?;
        client.add_to_data_definition(id, "PLANE ALTITUDE", "meters", DataType::Float64)?;

        client.request_data_on_sim_object(id, Period::Second, Condition::None, 0)?;

        Ok(())
    }
}

impl TryFrom<&NotificationData> for GpsData {
    type Error = SimConnectError;

    fn try_from(value: &NotificationData) -> Result<Self, Self::Error> {
        value.try_transmute::<GpsData>()
    }
}

fn main() -> Result<(), Box<dyn std::error::Error>> {
    let client = SimConnect::new("Simple Program");

    match client {
        Ok(mut client) => loop {
            let notification = client.get_next_dispatch()?;

            match notification {
                Some(Notification::Open) => {
                    println!("Open");

                    // The struct must be registered after the connection is successfully open
                    client.register_object::<GpsData>()?;
                }
                Some(Notification::Data(data)) => {
                    if let Ok(gps_data) = GpsData::try_from(&data) {
                        println!("GPS Data: {gps_data:?}");
                    }
                }
                _ => (),
            }

            // sleep for about a frame to reduce CPU usage
            std::thread::sleep(std::time::Duration::from_millis(16));
        },
        Err(e) => {
            println!("Error: {e:?}")
        }
    }

    Ok(())
}
